@@ -25,6 +25,18 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "mediastreamer2/msfilter.h"
 #include "mediastreamer2/mssndcard.h"
 
+//#define THREADED_VERSION
+
+/*in case of troubles with a particular driver, try incrementing ALSA_PERIOD_SIZE
+to 512, 1024, 2048, 4096...
+then try incrementing the number of periods*/
+#define ALSA_PERIODS 8
+#define ALSA_PERIOD_SIZE 256
+
+/*uncomment the following line if you have problems with an alsa driver
+having sound quality trouble:*/
+/*#define EPIPE_BUGFIX 1*/
+
 static MSSndCard * alsa_card_new(int id);
 static MSSndCard *alsa_card_duplicate(MSSndCard *obj);
 static MSFilter * ms_alsa_read_new(const char *dev);
@@ -47,8 +59,8 @@ static int alsa_set_params(snd_pcm_t *pcm_handle, int rw, int bits, int stereo,
 	uint exact_uvalue;
 	unsigned long exact_ulvalue;
 	int channels;
-	int periods=8;
-	int periodsize=256;
+	int periods=ALSA_PERIODS;
+	int periodsize=ALSA_PERIOD_SIZE;
 	int err;
 	int format;
 	
@@ -57,25 +69,25 @@ static int alsa_set_params(snd_pcm_t *pcm_handle, int rw, int bits, int stereo,
 	
 	/* Init hwparams with full configuration space */
 	if (snd_pcm_hw_params_any(pcm_handle, hwparams) < 0) {
-		ms_warning("alsa_set_params: Cannot configure this PCM device.\n");
+		ms_warning("alsa_set_params: Cannot configure this PCM device.");
 		return -1;
 	}
 	
 	if (snd_pcm_hw_params_set_access(pcm_handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED) < 0) {
-		ms_warning("alsa_set_params: Error setting access.\n");
+		ms_warning("alsa_set_params: Error setting access.");
 		return -1;
 	}
 	/* Set sample format */
 	format=SND_PCM_FORMAT_S16;
 	if (snd_pcm_hw_params_set_format(pcm_handle, hwparams, format) < 0) {
-		ms_warning("alsa_set_params: Error setting format.\n");
+		ms_warning("alsa_set_params: Error setting format.");
 		return -1;
 	}
 	/* Set number of channels */
 	if (stereo) channels=2;
 	else channels=1;
 	if (snd_pcm_hw_params_set_channels(pcm_handle, hwparams, channels) < 0) {
-		ms_warning("alsa_set_params: Error setting channels.\n");
+		ms_warning("alsa_set_params: Error setting channels.");
 		return -1;
 	}
 	/* Set sample rate. If the exact rate is not supported */
@@ -88,7 +100,7 @@ static int alsa_set_params(snd_pcm_t *pcm_handle, int rw, int bits, int stereo,
 	}
 	if (dir != 0) {
 		ms_warning("alsa_set_params: The rate %d Hz is not supported by your hardware.\n "
-		"==> Using %d Hz instead.\n", rate, exact_uvalue);
+		"==> Using %d Hz instead.", rate, exact_uvalue);
 	}
 	/* choose greater period size when rate is high */
 	periodsize=periodsize*(rate/8000);	
@@ -99,24 +111,26 @@ static int alsa_set_params(snd_pcm_t *pcm_handle, int rw, int bits, int stereo,
 	exact_ulvalue=periodsize;
 	dir=0;
 	if (snd_pcm_hw_params_set_period_size_near(pcm_handle, hwparams, &exact_ulvalue, &dir) < 0) {
-		ms_warning("alsa_set_params: Error setting period size.\n");
+		ms_warning("alsa_set_params: Error setting period size.");
 		return -1;
 	}
 	if (dir != 0) {
 		ms_warning("alsa_set_params: The period size %d is not supported by your hardware.\n "
-		"==> Using %d instead.\n", periodsize, (int)exact_ulvalue);
+		"==> Using %d instead.", periodsize, (int)exact_ulvalue);
 	}
+	ms_warning("alsa_set_params: periodsize:%d Using %d", periodsize, (int)exact_ulvalue);
 	periodsize=exact_ulvalue;
 	/* Set number of periods. Periods used to be called fragments. */ 
 	exact_uvalue=periods;
 	dir=0;
 	if (snd_pcm_hw_params_set_periods_near(pcm_handle, hwparams, &exact_uvalue, &dir) < 0) {
-		ms_warning("alsa_set_params: Error setting periods.\n");
+		ms_warning("alsa_set_params: Error setting periods.");
 		return -1;
 	}
+	ms_warning("alsa_set_params: period:%d Using %d", periods, exact_uvalue);
 	if (dir != 0) {
 		ms_warning("alsa_set_params: The number of periods %d is not supported by your hardware.\n "
-		"==> Using %d instead.\n", periods, exact_uvalue);
+		"==> Using %d instead.", periods, exact_uvalue);
 	}
 	/* Apply HW parameter settings to */
 	/* PCM device and prepare device  */
@@ -130,7 +144,9 @@ static int alsa_set_params(snd_pcm_t *pcm_handle, int rw, int bits, int stereo,
 		snd_pcm_sw_params_current(pcm_handle, swparams);
 		if ((err=snd_pcm_sw_params_set_start_threshold(pcm_handle, swparams,periodsize*2 ))<0){
 			ms_warning("alsa_set_params: Error setting start threshold:%s",snd_strerror(err));
-			return -1;
+		}
+		if ((err=snd_pcm_sw_params_set_stop_threshold(pcm_handle, swparams,periodsize*periods ))<0){
+			ms_warning("alsa_set_params: Error setting stop threshold:%s",snd_strerror(err));
 		}
 		if ((err=snd_pcm_sw_params(pcm_handle, swparams))<0){
 			ms_warning("alsa_set_params: Error setting SW params:%s",snd_strerror(err));
@@ -140,22 +156,88 @@ static int alsa_set_params(snd_pcm_t *pcm_handle, int rw, int bits, int stereo,
 	return 0;	
 }
 
+#ifdef EPIPE_BUGFIX
+static void alsa_fill_w (snd_pcm_t *pcm_handle)
+{
+	snd_pcm_hw_params_t *hwparams=NULL;
+	int channels;
+        snd_pcm_uframes_t buffer_size;
+	int buffer_size_bytes;
+	void *buffer;
+
+	/* Allocate the snd_pcm_hw_params_t structure on the stack. */
+	snd_pcm_hw_params_alloca(&hwparams);
+	snd_pcm_hw_params_current(pcm_handle, hwparams);
+
+	/* get channels */
+	snd_pcm_hw_params_get_channels (hwparams, &channels);
+
+	/* get buffer size */
+	snd_pcm_hw_params_get_buffer_size (hwparams, &buffer_size);
+
+	/* fill half */
+	buffer_size /= 2;
+
+	/* allocate buffer assuming 2 bytes per sample */
+	buffer_size_bytes = buffer_size * channels * 2;
+	buffer = alloca (buffer_size_bytes);
+	memset (buffer, 0, buffer_size_bytes);
+
+	/* write data */
+	snd_pcm_writei(pcm_handle, buffer, buffer_size);
+}
+#endif
+
 static snd_pcm_t * alsa_open_r(const char *pcmdev,int bits,int stereo,int rate)
 {
 	snd_pcm_t *pcm_handle;
 	int err;
 
+	ms_message("alsa_open_r: opening %s at %iHz, bits=%i, stereo=%i",pcmdev,rate,bits,stereo);
+
+
+#ifndef THREADED_VERSION
 	if (snd_pcm_open(&pcm_handle, pcmdev,SND_PCM_STREAM_CAPTURE,SND_PCM_NONBLOCK) < 0) {
-		ms_warning("alsa_card_open_r: Error opening PCM device %s\n",pcmdev );
+		ms_warning("alsa_open_r: Error opening PCM device %s",pcmdev );
 		return NULL;
 	}
-	if (alsa_set_params(pcm_handle,0,bits,stereo,rate)<0){
-		snd_pcm_close(pcm_handle);
+#else
+	/* want blocking mode for threaded version */
+	if (snd_pcm_open(&pcm_handle, pcmdev,SND_PCM_STREAM_CAPTURE,0) < 0) {
+		ms_warning("alsa_open_r: Error opening PCM device %s",pcmdev );
 		return NULL;
 	}
+#endif
+
+	{
+	struct timeval tv1;
+	struct timeval tv2;
+	struct timezone tz;
+	int diff = 0;
+	err = gettimeofday(&tv1, &tz);
+	while (1) { 
+		if (!(alsa_set_params(pcm_handle,0,bits,stereo,rate)<0)){
+			ms_message("alsa_open_r: Audio params set");
+			break;
+		}
+		if (!gettimeofday(&tv2, &tz) && !err) {
+			diff = ((tv2.tv_sec - tv1.tv_sec) * 1000000) + (tv2.tv_usec - tv1.tv_usec);
+		} else {
+			diff = -1;
+		}
+		if ((diff < 0) || (diff > 3000000)) { /* 3 secondes */
+			ms_error("alsa_open_r: Error setting params for more than 3 seconds");
+			snd_pcm_close(pcm_handle);
+			return NULL;
+		}
+		ms_warning("alsa_open_r: Error setting params (for %d micros)", diff);
+		usleep(200000);
+	}
+	}
+
 	err=snd_pcm_start(pcm_handle);
 	if (err<0){
-		ms_warning("Cannot start read pcm: %s", snd_strerror(err));
+		ms_warning("snd_pcm_start() failed: %s", snd_strerror(err));
 	}
 	return pcm_handle;
 }
@@ -165,22 +247,73 @@ static snd_pcm_t * alsa_open_w(const char *pcmdev,int bits,int stereo,int rate)
 	snd_pcm_t *pcm_handle;
 	
 	if (snd_pcm_open(&pcm_handle, pcmdev,SND_PCM_STREAM_PLAYBACK,SND_PCM_NONBLOCK) < 0) {
-		ms_warning("alsa_card_open_r: Error opening PCM device %s\n",pcmdev );
+		ms_warning("alsa_open_w: Error opening PCM device %s",pcmdev );
 		return NULL;
 	}
-	if (alsa_set_params(pcm_handle,1,bits,stereo,rate)<0){
-		snd_pcm_close(pcm_handle);
-		return NULL;
+	
+	{
+	struct timeval tv1;
+	struct timeval tv2;
+	struct timezone tz;
+	int diff = 0;
+	int err;
+	err = gettimeofday(&tv1, &tz);
+	while (1) { 
+		if (!(alsa_set_params(pcm_handle,1,bits,stereo,rate)<0)){
+			ms_message("alsa_open_w: Audio params set");
+			break;
+		}
+		if (!gettimeofday(&tv2, &tz) && !err) {
+			diff = ((tv2.tv_sec - tv1.tv_sec) * 1000000) + (tv2.tv_usec - tv1.tv_usec);
+		} else {
+			diff = -1;
+		}
+		if ((diff < 0) || (diff > 3000000)) { /* 3 secondes */
+			ms_error("alsa_open_w: Error setting params for more than 3 seconds");
+			snd_pcm_close(pcm_handle);
+			return NULL;
+		}
+		ms_warning("alsa_open_w: Error setting params (for %d micros)", diff);
+		usleep(200000);
 	}
+	}
+
 	return pcm_handle;
 }
 
+static int alsa_can_read(snd_pcm_t *dev)
+{
+	snd_pcm_sframes_t avail;
+	int err;
+
+	avail = snd_pcm_avail_update(dev);
+	if (avail < 0) {
+		ms_error("snd_pcm_avail_update: %s", snd_strerror(avail));	// most probably -EPIPE
+		/* overrun occured, snd_pcm_state() would return SND_PCM_STATE_XRUN
+		 FIXME: handle other error conditions*/
+		ms_error("*** alsa_can_read fixup, trying to recover");
+		snd_pcm_drain(dev); /* Ignore possible error, at least -EAGAIN.*/
+		err = snd_pcm_recover(dev, avail, 0);
+		if (err){ 
+			ms_error("snd_pcm_recover() failed with err %d: %s", err, snd_strerror(err));
+			return -1;
+		}
+		err = snd_pcm_start(dev);
+		if (err){ 
+			ms_error("snd_pcm_start() failed with err %d: %s", err, snd_strerror(err)); 
+			return -1; 
+		}
+		ms_message("Recovery done");
+	}
+	return avail;
+}
 
 static int alsa_read(snd_pcm_t *handle,unsigned char *buf,int nsamples)
 {
 	int err;
 	err=snd_pcm_readi(handle,buf,nsamples);
 	if (err<0) {
+		ms_warning("alsa_read: snd_pcm_readi() returned %i",err);
 		if (err==-EPIPE){
 			snd_pcm_prepare(handle);
 			err=snd_pcm_readi(handle,buf,nsamples);
@@ -201,13 +334,16 @@ static int alsa_write(snd_pcm_t *handle,unsigned char *buf,int nsamples)
 	if ((err=snd_pcm_writei(handle,buf,nsamples))<0){
 		if (err==-EPIPE){
 			snd_pcm_prepare(handle);
+#ifdef EPIPE_BUGFIX
+			alsa_fill_w (handle);
+#endif
 			err=snd_pcm_writei(handle,buf,nsamples);
 			if (err<0) ms_warning("alsa_card_write: Error writing sound buffer (nsamples=%i):%s",nsamples,snd_strerror(err));
 		}else if (err!=-EWOULDBLOCK){
 			ms_warning("alsa_card_write: snd_pcm_writei() failed:%s.",snd_strerror(err));
 		}
 	}else if (err!=nsamples) {
-		ms_warning("Only %i samples written instead of %i",err,nsamples);
+		ms_debug("Only %i samples written instead of %i",err,nsamples);
 	}
 	return err;
 }
@@ -256,7 +392,7 @@ static int get_mixer_element(snd_mixer_t *mixer,const char *name, MixerAction ac
 	elem=snd_mixer_first_elem(mixer);
 	while (elem!=NULL){
 		elemname=snd_mixer_selem_get_name(elem);
-		//g_message("Found alsa mixer element %s.",elemname);
+		//ms_message("Found alsa mixer element %s.",elemname);
 		if (strcmp(elemname,name)==0){
 			switch (action){
 				case CAPTURE:
@@ -266,7 +402,7 @@ static int get_mixer_element(snd_mixer_t *mixer,const char *name, MixerAction ac
 					newvol-=sndMixerPMin;
 					value=(100*newvol)/(sndMixerPMax-sndMixerPMin);
 					if (err<0) ms_warning("Could not get capture volume for %s:%s",name,snd_strerror(err));
-					//else ms_message("Succesfully get capture level for %s.",elemname);
+					//else ms_message("Successfully get capture level for %s.",elemname);
 					break;
 				}
 				break;
@@ -277,7 +413,7 @@ static int get_mixer_element(snd_mixer_t *mixer,const char *name, MixerAction ac
 					newvol-=sndMixerPMin;
 					value=(100*newvol)/(sndMixerPMax-sndMixerPMin);
 					if (err<0) ms_warning("Could not get playback volume for %s:%s",name,snd_strerror(err));
-					//else g_message("Succesfully get playback level for %s.",elemname);
+					//else ms_message("Successfully get playback level for %s.",elemname);
 					break;
 				}
 				break;
@@ -307,7 +443,7 @@ static void set_mixer_element(snd_mixer_t *mixer,const char *name, int level,Mix
 	
 	while (elem!=NULL){
 		elemname=snd_mixer_selem_get_name(elem);
-		//g_message("Found alsa mixer element %s.",elemname);
+		//ms_message("Found alsa mixer element %s.",elemname);
 		if (strcmp(elemname,name)==0){
 			switch(action){
 				case CAPTURE:
@@ -315,7 +451,7 @@ static void set_mixer_element(snd_mixer_t *mixer,const char *name, int level,Mix
 					snd_mixer_selem_get_capture_volume_range(elem, &sndMixerPMin, &sndMixerPMax);
 					newvol=(((sndMixerPMax-sndMixerPMin)*level)/100)+sndMixerPMin;
 					snd_mixer_selem_set_capture_volume_all(elem,newvol);
-					//g_message("Succesfully set capture level for %s.",elemname);
+					//ms_message("Successfully set capture level for %s.",elemname);
 					return;
 				}
 				break;
@@ -324,20 +460,20 @@ static void set_mixer_element(snd_mixer_t *mixer,const char *name, int level,Mix
 					snd_mixer_selem_get_playback_volume_range(elem, &sndMixerPMin, &sndMixerPMax);
 					newvol=(((sndMixerPMax-sndMixerPMin)*level)/100)+sndMixerPMin;
 					snd_mixer_selem_set_playback_volume_all(elem,newvol);
-					//g_message("Succesfully set playback level for %s.",elemname);
+					//ms_message("Successfully set playback level for %s.",elemname);
 					return;
 				}
 				break;
 				case CAPTURE_SWITCH:
 				if (snd_mixer_selem_has_capture_switch(elem)){
 					snd_mixer_selem_set_capture_switch_all(elem,level);
-					//g_message("Succesfully set capture switch for %s.",elemname);
+					//ms_message("Successfully set capture switch for %s.",elemname);
 				}
 				break;
 				case PLAYBACK_SWITCH:
 				if (snd_mixer_selem_has_playback_switch(elem)){
 					snd_mixer_selem_set_playback_switch_all(elem,level);
-					//g_message("Succesfully set capture switch for %s.",elemname);
+					//ms_message("Successfully set capture switch for %s.",elemname);
 				}
 				break;
 
@@ -439,9 +575,9 @@ static void alsa_card_uninit(MSSndCard *obj){
 	AlsaData *ad=(AlsaData*)obj->data;
 	if (ad->pcmdev!=NULL) ms_free(ad->pcmdev);
 	if (ad->mixdev!=NULL) ms_free(ad->mixdev);
+	ms_free(ad);
 }
 
-
 static void alsa_card_detect(MSSndCardManager *m){
 	int i;
 	for (i=-1;i<10;i++){
@@ -458,6 +594,8 @@ MSSndCardDesc alsa_card_desc={
 	.set_level=alsa_card_set_level,
 	.get_level=alsa_card_get_level,
 	.set_capture=alsa_card_set_source,
+	.set_control=NULL,
+	.get_control=NULL,
 	.create_reader=alsa_card_create_reader,
 	.create_writer=alsa_card_create_writer,
 	.uninit=alsa_card_uninit,
@@ -469,17 +607,44 @@ static MSSndCard *alsa_card_duplicate(MSSndCard *obj){
 	AlsaData* dcard=(AlsaData*)card->data;
 	AlsaData* dobj=(AlsaData*)obj->data;
 	card->name=ms_strdup(obj->name);
+	card->id=ms_strdup(obj->id);
 	dcard->pcmdev=ms_strdup(dobj->pcmdev);
 	dcard->mixdev=ms_strdup(dobj->mixdev);
+	return card;
+}
+
+MSSndCard * ms_alsa_card_new_custom(const char *pcmdev, const char *mixdev){
+	MSSndCard * obj;
+	AlsaData *ad;
+	obj=ms_snd_card_new(&alsa_card_desc);
+	ad=(AlsaData*)obj->data;
+	obj->name=ms_strdup(pcmdev);
+	ad->pcmdev=ms_strdup(pcmdev);
+	ad->mixdev=ms_strdup(mixdev);
 	return obj;
 }
 
+static unsigned int get_card_capabilities(const char *devname){
+	snd_pcm_t *pcm_handle;
+	unsigned int ret=0;
+	if (snd_pcm_open(&pcm_handle,devname,SND_PCM_STREAM_CAPTURE,SND_PCM_NONBLOCK)==0) {
+		ret|=MS_SND_CARD_CAP_CAPTURE;
+		snd_pcm_close(pcm_handle);
+	}
+	if (snd_pcm_open(&pcm_handle,devname,SND_PCM_STREAM_PLAYBACK,SND_PCM_NONBLOCK)==0) {
+		ret|=MS_SND_CARD_CAP_PLAYBACK;
+		snd_pcm_close(pcm_handle);
+	}
+	return ret;
+}
+
 static MSSndCard * alsa_card_new(int id)
 {
 	MSSndCard * obj;
 	char *name=NULL;
 	AlsaData *ad;
 	int err;
+	
 	if (id!=-1){
 		err=snd_card_get_name(id,&name);
 		if (err<0) {
@@ -500,8 +665,30 @@ static MSSndCard * alsa_card_new(int id)
 		pos2=pos1+strlen(pos1)-1;
 		for (; pos2>pos1 && *pos2==' '; pos2--) *pos2='\0';
 		obj->name=pos1;
-		ad->pcmdev=ms_strdup_printf("plughw:%i,0",id);
-		ad->mixdev=ms_strdup_printf("hw:%i",id);
+		ad->pcmdev=ms_strdup_printf("default:%i",id);
+		ad->mixdev=ms_strdup_printf("default:%i",id);
+		{
+			snd_mixer_t *mixer;
+			mixer = alsa_mixer_open(ad->mixdev);
+			if (mixer==NULL) {
+				ms_free(ad->mixdev);
+				ad->mixdev=ms_strdup_printf("hw:%i",id);
+			} else {
+				alsa_mixer_close(mixer);
+			}
+		}
+	}
+	/*check card capabilities: */
+	obj->capabilities=get_card_capabilities(ad->pcmdev);
+	if (obj->capabilities==0){
+		ms_warning("Strange, sound card %s does not seems to be capable of anything, retrying with plughw...",obj->name);
+		/*retry with plughw: this workarounds an alsa bug*/
+		ms_free(ad->pcmdev);
+		ad->pcmdev=ms_strdup_printf("plughw:%i",id);
+		obj->capabilities=get_card_capabilities(ad->pcmdev);
+		if (obj->capabilities==0){
+			ms_warning("Strange, sound card %s seems totally unusable.",obj->name);
+		}
 	}
 	free(name);
 	/*ms_message("alsa device %s found",obj->name);*/
@@ -513,6 +700,14 @@ struct _AlsaReadData{
 	snd_pcm_t *handle;
 	int rate;
 	int nchannels;
+
+#ifdef THREADED_VERSION
+	ms_thread_t thread;
+	ms_mutex_t mutex;
+	MSBufferizer * bufferizer;
+	bool_t read_started;
+	bool_t write_started;
+#endif
 };
 
 typedef struct _AlsaReadData AlsaReadData;
@@ -524,39 +719,143 @@ void alsa_read_init(MSFilter *obj){
 	ad->rate=8000;
 	ad->nchannels=1;
 	obj->data=ad;
+
+#ifdef THREADED_VERSION
+	ad->read_started=FALSE;
+	ad->write_started=FALSE;
+	ad->bufferizer=ms_bufferizer_new();
+	ms_mutex_init(&ad->mutex,NULL);
+	ad->thread=0;
+#endif
 }
 
+#ifdef THREADED_VERSION
+
+static void * alsa_write_thread(void *p){
+	AlsaReadData *ad=(AlsaReadData*)p;
+	int samples=(160*ad->rate)/8000;
+	int err;
+	int count=0;
+	mblk_t *om=NULL;
+	struct timeval timeout;
+	if (ad->handle==NULL && ad->pcmdev!=NULL){
+		ad->handle=alsa_open_r(ad->pcmdev,16,ad->nchannels==2,ad->rate);
+	}
+	if (ad->handle==NULL) return NULL;
+
+	while (ad->read_started)
+	  {
+	    count = alsa_can_read(ad->handle,samples);
+	    if (count==24)
+	      { /* keep this value for this driver */ }
+	    else if (count<=0)
+	      {
+		count = samples;
+	      }
+	    else if (count>0)
+	      {
+		//ms_warning("%i count", count);
+		//count = samples;
+	      }
+
+	    int size=count*2;
+	    om=allocb(size,0);
+
+	    if ((err=alsa_read(ad->handle,om->b_wptr,count))<=0)
+	      {
+		ms_warning("nothing to read");
+		//ms_warning("Fail to read samples %i", count);
+		freemsg(om); /* leak fixed */
+		continue;
+	      }
+	    //ms_warning(" read %i", err);
+	    
+	    size=err*2;
+	    om->b_wptr+=size;
+
+	    ms_mutex_lock(&ad->mutex);
+	    ms_bufferizer_put(ad->bufferizer,om);
+	    ms_mutex_unlock(&ad->mutex);
+
+	    if (count==24)
+	      {
+		timeout.tv_sec = 0;
+		timeout.tv_usec = 2000;
+		select(0, 0, NULL, NULL, &timeout );
+	      }
+	    else
+	      {
+		/* select will be less active than locking on "read" */
+		timeout.tv_sec = 0;
+		timeout.tv_usec = 5000;
+		select(0, 0, NULL, NULL, &timeout );
+	      }
+	  }
+
+	if (ad->handle!=NULL) snd_pcm_close(ad->handle);
+	ad->handle=NULL;
+	return NULL;
+}
+
+static void alsa_start_r(AlsaReadData *d){
+	if (d->read_started==FALSE){
+		d->read_started=TRUE;
+		ms_thread_create(&d->thread,NULL,alsa_write_thread,d);
+	}else d->read_started=TRUE;
+}
+
+static void alsa_stop_r(AlsaReadData *d){
+	d->read_started=FALSE;
+	if (d->thread!=0)
+	  {
+	    ms_thread_join(d->thread,NULL);
+	    d->thread=0;
+	  }
+}
+#endif
+
+#ifdef THREADED_VERSION
+void alsa_read_preprocess(MSFilter *obj){
+	AlsaReadData *ad=(AlsaReadData*)obj->data;
+	alsa_start_r(ad);
+}
+#endif
+
 void alsa_read_postprocess(MSFilter *obj){
 	AlsaReadData *ad=(AlsaReadData*)obj->data;
+#ifdef THREADED_VERSION
+	alsa_stop_r(ad);
+#endif
 	if (ad->handle!=NULL) snd_pcm_close(ad->handle);
 	ad->handle=NULL;
 }
 
 void alsa_read_uninit(MSFilter *obj){
 	AlsaReadData *ad=(AlsaReadData*)obj->data;
+#ifdef THREADED_VERSION
+	alsa_stop_r(ad);
+#endif
 	if (ad->pcmdev!=NULL) ms_free(ad->pcmdev);
 	if (ad->handle!=NULL) snd_pcm_close(ad->handle);
+#ifdef THREADED_VERSION
+	ms_bufferizer_destroy(ad->bufferizer);
+	ms_mutex_destroy(&ad->mutex);
+#endif
 	ms_free(ad);
 }
 
-static bool_t alsa_can_read(snd_pcm_t *dev, int frames){
-	return snd_pcm_avail_update(dev)>=frames;
-}
-
+#ifndef THREADED_VERSION
 void alsa_read_process(MSFilter *obj){
 	AlsaReadData *ad=(AlsaReadData*)obj->data;
-#ifdef AMD_HACK
-	int samples=(160*ad->rate)/8000;
-#else
 	int samples=(128*ad->rate)/8000;
-#endif
 	int err;
 	mblk_t *om=NULL;
 	if (ad->handle==NULL && ad->pcmdev!=NULL){
 		ad->handle=alsa_open_r(ad->pcmdev,16,ad->nchannels==2,ad->rate);
 	}
 	if (ad->handle==NULL) return;
-	while (alsa_can_read(ad->handle,samples)){
+	while (alsa_can_read(ad->handle)>=samples){
+	  
 		int size=samples*2;
 		om=allocb(size,0);
 		if ((err=alsa_read(ad->handle,om->b_wptr,samples))<=0) {
@@ -568,11 +867,34 @@ void alsa_read_process(MSFilter *obj){
 		om->b_wptr+=size;
 		/*ms_message("alsa_read_process: Outputing %i bytes",size);*/
 		ms_queue_put(obj->outputs[0],om);
-#ifdef AMD_HACK
-		break;
-#endif
 	}
 }
+#endif
+
+#ifdef THREADED_VERSION
+void alsa_read_process(MSFilter *obj){
+	AlsaReadData *ad=(AlsaReadData*)obj->data;
+	mblk_t *om=NULL;
+	int samples=(160*ad->rate)/8000;
+
+	ms_mutex_lock(&ad->mutex);
+	while (ms_bufferizer_get_avail(ad->bufferizer)>=samples*2){
+	  
+	  om=allocb(samples*2,0);
+	  ms_bufferizer_read(ad->bufferizer,om->b_wptr,samples*2);	  
+	  om->b_wptr+=samples*2;
+	  /*ms_message("alsa_read_process: Outputing %i bytes",size);*/
+	  ms_queue_put(obj->outputs[0],om);
+	}
+	ms_mutex_unlock(&ad->mutex);
+}
+#endif
+
+static int alsa_read_get_sample_rate(MSFilter *obj, void *param){
+	AlsaReadData *ad=(AlsaReadData*)obj->data;
+	*((int*)param)=ad->rate;
+	return 0;
+}
 
 static int alsa_read_set_sample_rate(MSFilter *obj, void *param){
 	AlsaReadData *ad=(AlsaReadData*)obj->data;
@@ -587,19 +909,23 @@ static int alsa_read_set_nchannels(MSFilter *obj, void *param){
 }
 
 MSFilterMethod alsa_read_methods[]={
+	{MS_FILTER_GET_SAMPLE_RATE,	alsa_read_get_sample_rate},
 	{MS_FILTER_SET_SAMPLE_RATE, alsa_read_set_sample_rate},
-	{MS_FILTER_SET_SAMPLE_RATE, alsa_read_set_nchannels},
+	{MS_FILTER_SET_NCHANNELS, alsa_read_set_nchannels},
 	{0,NULL}
 };
 
 MSFilterDesc alsa_read_desc={
 	.id=MS_ALSA_READ_ID,
 	.name="MSAlsaRead",
-	.text="Alsa sound source",
+	.text=N_("Alsa sound source"),
 	.category=MS_FILTER_OTHER,
 	.ninputs=0,
 	.noutputs=1,
 	.init=alsa_read_init,
+#ifdef THREADED_VERSION
+	.preprocess=alsa_read_preprocess,
+#endif
 	.process=alsa_read_process,
 	.postprocess=alsa_read_postprocess,
 	.uninit=alsa_read_uninit,
@@ -637,6 +963,12 @@ void alsa_write_uninit(MSFilter *obj){
 	ms_free(ad);
 }
 
+static int alsa_write_get_sample_rate(MSFilter *obj, void *data){
+	AlsaWriteData *ad=(AlsaWriteData*)obj->data;
+	*((int*)data)=ad->rate;
+	return 0;
+}
+
 int alsa_write_set_sample_rate(MSFilter *obj, void *data){
 	int *rate=(int*)data;
 	AlsaWriteData *ad=(AlsaWriteData*)obj->data;
@@ -659,6 +991,9 @@ void alsa_write_process(MSFilter *obj){
 	int err;
 	if (ad->handle==NULL && ad->pcmdev!=NULL){
 		ad->handle=alsa_open_w(ad->pcmdev,16,ad->nchannels==2,ad->rate);
+#ifdef EPIPE_BUGFIX
+		alsa_fill_w (ad->pcmdev);
+#endif
 	}
 	if (ad->handle==NULL) {
 		ms_queue_flush(obj->inputs[0]);
@@ -668,7 +1003,9 @@ void alsa_write_process(MSFilter *obj){
 		while((size=im->b_wptr-im->b_rptr)>0){
 			samples=size/(2*ad->nchannels);
 			err=alsa_write(ad->handle,im->b_rptr,samples);
-			if (err>0) im->b_rptr+=err*(2*ad->nchannels);
+			if (err>0) {
+				im->b_rptr+=err*(2*ad->nchannels);
+			}
 			else break;
 		}
 		freemsg(im);
@@ -676,6 +1013,7 @@ void alsa_write_process(MSFilter *obj){
 }
 
 MSFilterMethod alsa_write_methods[]={
+	{MS_FILTER_GET_SAMPLE_RATE,	alsa_write_get_sample_rate},
 	{MS_FILTER_SET_SAMPLE_RATE, alsa_write_set_sample_rate},
 	{MS_FILTER_SET_NCHANNELS, alsa_write_set_nchannels},
 	{0,NULL}
@@ -684,7 +1022,7 @@ MSFilterMethod alsa_write_methods[]={
 MSFilterDesc alsa_write_desc={
 	.id=MS_ALSA_WRITE_ID,
 	.name="MSAlsaWrite",
-	.text="Alsa sound output",
+	.text=N_("Alsa sound output"),
 	.category=MS_FILTER_OTHER,
 	.ninputs=1,
 	.noutputs=0,
